#!/usr/bin/perl -w

=head1 NAME

B<POWER::Feed> - 

=head1 SYNOPSIS

 use POWER::Feed @dirs;

 @feed = (
    { feed => "Web::www_av_com", query => "perl" },
    { feed => "SportBet::www_betandwin_com" },
    );
 @RESULT = Feed(\@feed, $timeout, $DEBUG);

=head1 DESCRIPTION

  :  User-agent, Referer,
 Cookies, ,  POST  .

  :     
      ,  
     . , 
  ,     .
         
 ,     .

       / 
.

/ :    ,
 /  ,     
 /.

=head1 FUNCTIONS

=over 4

=cut

package POWER::Feed;

use POWER::lib qw(time);
use POWER::LOG;
use POWER::Multi::GET 2.00;
use POWER::Tree::AnyValue;

use strict;
our ($DEBUG, $NODE, @RESULT, %URL, $URL, $HTML, %Data, %Feed);	#  local  Feed()
our (%TIME);

our $VERSION = '0.10';

my @Path;				    #     feeds
sub import {
    no strict;
    *{"${\scalar caller}::$_"} = \&{$_} for qw(Feed load_feed);
    shift;
    my %seen;
    @Path = grep {!$seen{$_}++} @Path, @_,
	grep {-d} map {$_.'/POWER/Feed'} @INC;
}

our $LOG	    = $POWER::LOG::DEFAULT;
my $ROOTPKG	    = __PACKAGE__.'::';	    #     feeds
my %FEED	    = ();		    #   feeds
our %DTD	    = ();		    #  RESULT()   
our $HandlerTimeout = undef;		    #    

=item B<load_feed>(@dir)

         
feeds      feed   .

     .share        feed
  feed.

      .

        feed,  
    ,      eval.

        \W   _ .

=cut

sub load_feed {
    my %feeds;
    for (@_) {
	my $path = $_;
	$path =~ s{//+|/*$}{/}g;
	(my $pkg = $path) =~ s!/!::!g;
	_load_feed($_, $pkg) for grep {-d} map { $_.'/'.$path } @Path;
	$feeds{$_->{feed}} =
	    {path=>$_->{path}, file=>$_->{file}, type=>$_->{type}}
	    for grep { $_->{path} eq $pkg } values %FEED;
    }
    return \%feeds;
}
{ my %loaded_pkg;
sub _load_feed {
    my ($dir, $pkg) = @_;
    $pkg ||= '';
    my $share= Cat($dir."/".".share");
    opendir my $D, $dir or die $!; 
    for my $item (grep {m/^[^.]/} readdir $D) {
	my $path = $dir."/".$item;
	$item =~ s/\W/_/g;
	my $pkg  = $pkg.$item;
	if (-d $path) {
	    _load_feed($path, $pkg.'::');
	} elsif (-f $path and not $loaded_pkg{$pkg}++) {
	    Eval "package $ROOTPKG$pkg;\n" .
		"\n#line 1 \".share\"\n" . $share .
		"\n#line 1 \"$item\"\n"  . Cat($path);
	    die $@ if $@;
	}
    }
}}

=item B<Add_Feed>($type, $get [, \%Data])

    Feed.

 $get      ,    , 
" url". :
    ""
    " url"
    " url?get_params"
    " url post_params"
   ,        feed-.

        feed   
feed, pkg, path, file, type, handler, data, get.

=cut

sub Add_Feed {
    my ($pkg, $type, $get, $data) = (scalar(caller), @_);
    return unless $get;
    $data ||= {};
    my ($handler, $url, $post_data) = split " ", $get, 3;
    $pkg =~ /^\Q$ROOTPKG\E(.*::)?([^:]+)$/;
    $FEED{$1.$2.'/'.$type} = {
	feed=>$1.$2.'/'.$type, path=>$1, file=>$2, type=>$type,
	pkg=>$pkg, handler=>$handler, data=>{%$data},
	get=>($url ? {type=>'URL', handler=>$handler, url=>$url,
		      post_data=>$post_data}
		   : {type=>'HANDLER', handler=>$handler} ) };
}

=item B<>()



=cut

{ my $ID;
sub _subst {
    DEBUG_STRING("ERROR: \$Data{$_[0]} not exists\n") unless exists $Data{$_[0]};
    $Data{$_[0]};
}
sub GET {
    my ($get, $data) = @_;
    ref eq 'HASH' or $_={} for $get, $data;
  my $DIE; local $SIG{ALRM} = sub { $DIE=1 };
    if ($get->{type} eq 'HANDLER') {
	local $NODE = $DEBUG->add_value($NODE, ['HANDLER', $get->{handler}]);
	local %Data = %$data;
	runHandler($Feed{pkg}."::".$get->{handler});
    } elsif ($get->{type} eq 'URL') {
	#  ^~...~^  #~...~#  url/post_data    %Data.
	for (qw(url post_data)) {
	    while ( 1 ) {
		$get->{$_} =~ s/\G#~(.*?)~#/_subst($1)/gcse
	     or $get->{$_} =~ s/\G\^~(.*?)~\^/EncUri(_subst($1))/gcse
	     or $get->{$_} =~  /\G(.*?)(?=#~|\^~)/gcs
	     or last;
	    }
	}
	++$ID; $URL{$ID} = {
	    url		    => abs_url($get->{url}, $URL->{_url}),
	    add_to_header   => (exists $get->{add_to_header}
		? $get->{add_to_header}
		: 'Referer: '.($URL->{_url} || $get->{url})."\r\n"),
	    post_data	    => $get->{post_data},
	    cookie	    => $get->{cookie} || $URL->{_cookie},
	    (exists $get->{proxy} ? (proxy => $get->{proxy}) : ()),
	    downloaded	    => \&Feed_Downloaded,
	    (exists $get->{only_header} ? (only_header=>$get->{only_header}) : ()),
	    ID		    => $ID,
	    Handler	    => $Feed{pkg}.'::'.$get->{handler},
	    Feed	    => {%Feed},
	    Data	    => {%$data},
	    NODE	    => $DEBUG->add_value($NODE, undef),
	    };
	$TIME{$Feed{file}}{$Feed{type}}{start} ||= time();
	$TIME{$Feed{file}}{$Feed{type}}{count}++;
	$TIME{$Feed{file}}{$Feed{type}}{queue}++;
    } else {
	DEBUG_STRING("ERROR: unknown GET() type='$get->{type}'\n");
    }
  die "alarm\n" if $DIE;
}}

sub URL {
    my ($handler, $url, $post) = split " ", shift, 3;
    return unless $url;
    GET({type=>'URL', handler=>$handler, url=>$url, post_data=>$post}, {@_});
}

=item B<>()



=cut

sub RESULT {
  my $DIE; local $SIG{ALRM} = sub { $DIE=1 };
    my @res = map { {%Feed, %$_} } $DTD{$Feed{type}}->(@_);
    $DEBUG->add_value($NODE, ["RESULT", $_]) for @res;
    push @RESULT, @res;
  die "alarm\n" if $DIE;
}

=item B<>()



=cut

sub DEBUG_STRING {
  my $DIE; local $SIG{ALRM} = sub { $DIE=1 };
    $DEBUG->add_value($NODE, ["STRING", "@_"]);
  die "alarm\n" if $DIE;
}

=item B<>()



=cut

sub Feed {
    my ($feeds, $timeout) = @_;
    ($DEBUG, $NODE) = (POWER::Tree::AnyValue->root(), '/');
    (@RESULT,%URL,$URL,$HTML,%Data,%Feed,%TIME) = ();
    for (@$feeds) {
	my ($feed, $data) = @$_;
	$DEBUG->add_value($NODE, ['STRING',"Bad feed: `$feed'"]), next
	    unless exists $FEED{$feed};
	%Feed = %{$FEED{$feed}};
	%Data = (%{$Feed{data}}, %{$data||{}});
	GET($Feed{get}, \%Data);
    }
    POWER::Multi::GET::Download(%URL, undef, $timeout);
    $_[2] = $DEBUG if @_ > 2;
    return @RESULT;
}

sub Feed_Downloaded {
    $URL    = delete $URL{$_[0]->{ID}};
    $HTML   = $URL->{_content};
    %Data   = %{$URL->{Data}||{}};
    %Feed   = %{$URL->{Feed}||{}};
    $NODE   = $URL->{NODE};
    @{$URL}{"_ok","_err"} = ("ERR","Handler not found") 
	if $URL->{_ok} eq "OK" and not exists &{$URL->{Handler}};
    DEBUG_URL($URL);
    my $t = $TIME{$Feed{file}}{$Feed{type}};
    $t->{size} += length($URL->{_content});
    my $t0 = time();
    runHandler($URL->{Handler}) if $URL->{_ok} eq "OK";
    $t->{handler} += time()-$t0;
    $t->{queue}--;
    $LOG->DEBUG("fetch=%3d urls=%2d size=%4dKB type=%s file=%s\n",
	(time() - $t->{start}) - $t->{handler}, $t->{count}, $t->{size}/1024,
	$Feed{type}, $Feed{file}) unless $t->{queue};
    @{$URL}{"_ok","_err"} = ("ERR","Handler timed out")
	if $URL->{_ok} eq "OK" and $@ =~ /alarm/;
    $DEBUG->value($NODE, ["URL", {
	ID	    => $URL->{ID},
	url	    => $URL->{url},
	_ok	    => $URL->{_ok},
	_err	    => $URL->{_err},
	_redirect   => $URL->{_redirect},
	_size	    => length($URL->{_content}),
	Handler	    => map({local $_=$_;s/^\Q$ROOTPKG\E//;$_} $URL->{Handler}),
	Feed	    => {%Feed},
	}]);
}

sub runHandler {
    eval {
	local $SIG{__DIE__} = "IGNORE";
	#  die   ,   URL(), RESULT() 
	# DEBUG_STRING()  die     - 
	# die    POWER::Tree,    
	#  ,  tree2list    
	local $SIG{ALRM} = sub { die "alarm\n" };
	alarm($HandlerTimeout) if $HandlerTimeout;
        no strict;
	#    ,   
	#   next/last/redo  {Handler}
	{ &{ $_[0] } }
	alarm(0);
    };
}

=item B<>()



=cut

sub DEBUG_URL {
}

=back

=head1 VARIABLES

=over 4

=item B<>



=back

=head1 HISTORY

 0.10		  17 14:56:36 EET 2005
	- INCOMPATIBLE! feed-   
	   use POWER::Feed;
	- INCOMPATIBLE!    %POWER::Feed::FEED
	     load_feed()
	- INCOMPATIBLE!   Feed() Σ,  
	     feed  
	- INCOMPATIBLE!  url    ^~...~^ 
	  #~...~#,       %Data
	- Add_Feed()   %Data
	- Add_Feed()     url
	-   GET()     ( url
	    URL(), url   ,
	      url)
	-    feed-   
	   feeds  load_feed()

 0.08		  16 16:58:47 EET 2005
	-   POWER::LOG
	-        

 0.07		  15 03:08:40 EEST 2003
	-    .share    ,
	        
	-  ,       POWER::Feed, 
	          .share
	-         
	    

 0.05		   8 21:32:32 EEST 2003
	-   Regexp::Common::Fast,  -
	     ,     
	-    POWER::Feed   

 0.04		  27 09:19:06 EET 2003
	-    / :
	    die("alarm\n")    RESULT(),
	  URL()  DEBUG_STRING()

 0.03		  12 15:37:31 EET 2003
	-    /  
	  $POWER::Feed::HandlerTimeout = $seconds; # 0 = disable
	-      next/last/redo
	     

 0.02		   7 23:06:28 EET 2002
	-  RESULT()    

 0.01		  16 16:55:41 EEST 2002
	- initial release

=head1 AUTHOR

Alex Efros <powerman@asdfGroup.com>
Nikita Savin <nikita@asdfGroup.com>

=cut

1;
